from __future__ import annotations


def generate_tabular_overloads(func_name: str) -> str:
    """Generates overloads for a function with the specified engines and lazy options."""
    overloads = []
    ENGINES = ["pandas", "polars", "dask"]
    LAZY_OPTIONS = [False, True]

    for engine in ENGINES:
        for lazy in LAZY_OPTIONS:
            if engine == "pandas" and not lazy:
                return_type = "_pd.DataFrame"
            elif engine == "polars":
                return_type = "_pl.LazyFrame" if lazy else "_pl.DataFrame"
            elif engine == "dask" and lazy:
                return_type = "_dd.DataFrame"
            else:
                continue

            # Build the overload definition using .format
            overload_def = """\
@overload
def {func_name}(
    engine: {engine_literal},
    *,
    engine_kwargs: dict[str, Any] | None = None,
    lazy: {lazy_literal} = {lazy_repr},
) -> {return_type}: ...""".format(
                func_name=func_name,
                engine_literal=f'Literal["{engine}"]' if engine else "None",
                lazy_literal=f"Literal[{lazy}]",
                lazy_repr=lazy,
                return_type=return_type,
            )
            overloads.append(overload_def)

    return "\n".join(overloads)


def generate_gridded_overloads(func_name: str) -> str:
    """Generates overloads for a function with the specified engines and lazy options."""
    overloads = []
    ENGINES = ["xarray"]
    LAZY_OPTIONS = [False]

    for engine in ENGINES:
        for lazy in LAZY_OPTIONS:
            if engine is None:
                return_type = "" if lazy else "_xr.Dataset"
            elif engine == "xarray":
                return_type = "_xr.Dataset"
            else:
                continue

            # Build the overload definition using .format
            overload_def = """\
@overload
def {func_name}(
    engine: {engine_literal},
    *,
    engine_kwargs: dict[str, Any] | None = None,
    # lazy: {lazy_literal} = {lazy_repr},
) -> {return_type}: ...""".format(
                func_name=func_name,
                engine_literal=f'Literal["{engine}"]' if engine else "None",
                lazy_literal=f"Literal[{lazy}]",
                lazy_repr=lazy,
                return_type=return_type,
            )
            overloads.append(overload_def)

    return "\n".join(overloads)


synthetic_clusters = """@overload
def synthetic_clusters(
    engine: Literal["pandas"],
    *,
    lazy: Literal[False] = False,
    total_points: int = 1_000_000,
) -> _pd.DataFrame: ...
@overload
def synthetic_clusters(
    engine: Literal["polars"],
    *,
    lazy: Literal[False] = False,
    total_points: int = 1_000_000,
) -> _pl.DataFrame: ...
@overload
def synthetic_clusters(
    engine: Literal["polars"],
    *,
    lazy: Literal[True] = True,
    total_points: int = 1_000_000,
) -> _pl.LazyFrame: ...
@overload
def synthetic_clusters(
    engine: Literal["dask"],
    *,
    lazy: bool = False,
    total_points: int = 1_000_000,
) -> _dd.DataFrame: ...
"""
landsat_rgb = """@overload
def landsat_rgb(
    engine: Literal["rioxarray"],
    *,
    engine_kwargs: dict[str, Any] | None = None,
) -> _xr.Dataset: ...
"""
penguins_rgba = """@overload
def penguins_rgba(
    engine: Literal["xarray"],
) -> _xr.Dataset: ...
"""
us_states = """@overload
def us_states(
    engine: Literal["geopandas"],
    *,
    engine_kwargs: dict[str, Any] | None = None,
) -> _gpd.GeoDataFrame: ...
"""


def custom() -> str:
    return "\n".join(  # noqa: FLY002
        [
            synthetic_clusters.strip(),
            landsat_rgb.strip(),
            us_states.strip(),
        ]
    )


def main():
    header = """\
# !!! THIS FILE IS AUTOGENERATED !!!
# For changes see scripts/generate_typehints.py

from __future__ import annotations

from typing import Any, Literal, overload

import dask.dataframe as _dd
import geopandas as _gpd
import pandas as _pd
import polars as _pl
import xarray as _xr

__version__: str
__all__: tuple[str, ...]
"""

    return "\n".join(
        [
            header,
            generate_tabular_overloads("penguins"),
            generate_tabular_overloads("earthquakes"),
            generate_tabular_overloads("apple_stocks"),
            generate_tabular_overloads("stocks"),
            generate_tabular_overloads("nyc_taxi_remote"),
            generate_gridded_overloads("air_temperature"),
            custom(),
            "",
        ]
    )


if __name__ == "__main__":
    with open("src/hvsampledata/__init__.pyi", "w") as f:
        f.write(main())



from __future__ import annotations

import pytest

import hvsampledata as hvs
from hvsampledata._util import _EAGER_TABULAR_LOOKUP, _LAZY_TABULAR_LOOKUP

datasets = [
    hvs.apple_stocks,
    hvs.earthquakes,
    hvs.synthetic_clusters,
    hvs.penguins,
    hvs.stocks,
]


@pytest.mark.parametrize("dataset", datasets)
@pytest.mark.parametrize("engine", list(_EAGER_TABULAR_LOOKUP))
def test_eager_load(dataset, engine):
    pytest.importorskip(engine)
    df = dataset(engine=engine)
    if engine == "pandas":
        import pandas as pd

        assert isinstance(df, pd.DataFrame)
    elif engine == "polars":
        import polars as pl

        assert isinstance(df, pl.DataFrame)
    else:
        msg = "Not valid engine"
        raise ValueError(msg)


@pytest.mark.parametrize("dataset", datasets)
@pytest.mark.parametrize("engine", list(_LAZY_TABULAR_LOOKUP))
def test_lazy_load(dataset, engine):
    pytest.importorskip(engine)
    df = dataset(engine=engine, lazy=True)
    if engine == "polars":
        import polars as pl

        assert isinstance(df, pl.LazyFrame)
    elif engine == "dask":
        import dask.dataframe as dd

        assert isinstance(df, dd.DataFrame)
    else:
        msg = "Not valid engine"
        raise ValueError(msg)


@pytest.mark.parametrize("engine", list(_EAGER_TABULAR_LOOKUP))
def test_penguins_schema(engine):
    pytest.importorskip(engine)
    df = hvs.penguins(engine=engine)
    if engine == "pandas":
        import numpy as np
        import pandas as pd

        expected_dtypes = pd.Series(
            {
                "species": np.dtype("O"),
                "island": np.dtype("O"),
                "bill_length_mm": np.dtype("float64"),
                "bill_depth_mm": np.dtype("float64"),
                "flipper_length_mm": np.dtype("float64"),
                "body_mass_g": np.dtype("float64"),
                "sex": np.dtype("O"),
                "year": np.dtype("int64"),
            }
        )
        pd.testing.assert_series_equal(df.dtypes, expected_dtypes)
    elif engine == "polars":
        import polars as pl

        assert df.schema == {
            "species": pl.String,
            "island": pl.String,
            "bill_length_mm": pl.Float64,
            "bill_depth_mm": pl.Float64,
            "flipper_length_mm": pl.Int64,
            "body_mass_g": pl.Int64,
            "sex": pl.String,
            "year": pl.Int64,
        }
    else:
        msg = "Not valid engine"
        raise ValueError(msg)


@pytest.mark.parametrize("engine", list(_LAZY_TABULAR_LOOKUP))
def test_penguins_schema_lazy(engine):
    pytest.importorskip(engine)
    df = hvs.penguins(engine=engine, lazy=True)
    if engine == "dask":
        import numpy as np
        import pandas as pd

        expected_dtypes = pd.Series(
            {
                "species": pd.StringDtype("pyarrow"),
                "island": pd.StringDtype("pyarrow"),
                "bill_length_mm": np.dtype("float64"),
                "bill_depth_mm": np.dtype("float64"),
                "flipper_length_mm": np.dtype("float64"),
                "body_mass_g": np.dtype("float64"),
                "sex": pd.StringDtype("pyarrow"),
                "year": np.dtype("int64"),
            }
        )
        pd.testing.assert_series_equal(df.dtypes, expected_dtypes)
    elif engine == "polars":
        import polars as pl

        assert df.collect_schema() == {
            "species": pl.String,
            "island": pl.String,
            "bill_length_mm": pl.Float64,
            "bill_depth_mm": pl.Float64,
            "flipper_length_mm": pl.Int64,
            "body_mass_g": pl.Int64,
            "sex": pl.String,
            "year": pl.Int64,
        }
    else:
        msg = "Not valid engine"
        raise ValueError(msg)


@pytest.mark.parametrize("engine", list(_EAGER_TABULAR_LOOKUP))
def test_eager_load_earthquakes(engine):
    pytest.importorskip(engine)
    df = hvs.earthquakes(engine=engine)
    if engine == "pandas":
        import pandas as pd

        assert isinstance(df, pd.DataFrame)
    elif engine == "polars":
        import polars as pl

        assert isinstance(df, pl.DataFrame)
    else:
        msg = "Not valid engine"
        raise ValueError(msg)


@pytest.mark.parametrize("engine", list(_LAZY_TABULAR_LOOKUP))
def test_lazy_load_earthquake(engine):
    pytest.importorskip(engine)
    df = hvs.earthquakes(engine=engine, lazy=True)
    if engine == "polars":
        import polars as pl

        assert isinstance(df, pl.LazyFrame)
    elif engine == "dask":
        import dask.dataframe as dd

        assert isinstance(df, dd.DataFrame)
    else:
        msg = "Not valid engine"
        raise ValueError(msg)


@pytest.mark.parametrize("engine", list(_EAGER_TABULAR_LOOKUP))
def test_earthquakes_schema(engine):
    pytest.importorskip(engine)
    df = hvs.earthquakes(engine=engine)
    if engine == "pandas":
        import numpy as np
        import pandas as pd

        expected_dtypes = pd.Series(
            {
                "time": np.dtype("datetime64[ns]"),
                "lat": np.dtype("float64"),
                "lon": np.dtype("float64"),
                "depth": np.dtype("float64"),
                "depth_class": pd.CategoricalDtype(
                    categories=["Shallow", "Intermediate", "Deep"], ordered=True
                ),
                "mag": np.dtype("float64"),
                "mag_class": pd.CategoricalDtype(
                    categories=["Light", "Moderate", "Strong", "Major"], ordered=True
                ),
                "place": np.dtype("O"),
            }
        )
        pd.testing.assert_series_equal(df.dtypes, expected_dtypes)
    elif engine == "polars":
        import polars as pl

        expected_schema = {
            "time": pl.Datetime(time_unit="us", time_zone=None),
            "lat": pl.Float64,
            "lon": pl.Float64,
            "depth": pl.Float64,
            "depth_class": pl.Enum(categories=["Shallow", "Intermediate", "Deep"]),
            "mag": pl.Float64,
            "mag_class": pl.Enum(categories=["Light", "Moderate", "Strong", "Major"]),
            "place": pl.String,
        }
        assert df.schema == expected_schema
    else:
        msg = "Not valid engine"
        raise ValueError(msg)


@pytest.mark.parametrize("engine", list(_LAZY_TABULAR_LOOKUP))
def test_earthquakes_schema_lazy(engine):
    pytest.importorskip(engine)
    df = hvs.earthquakes(engine=engine, lazy=True)
    if engine == "dask":
        import numpy as np
        import pandas as pd

        expected_dtypes = pd.Series(
            {
                "time": np.dtype("datetime64[ns]"),
                "lat": np.dtype("float64"),
                "lon": np.dtype("float64"),
                "depth": np.dtype("float64"),
                "depth_class": pd.CategoricalDtype(
                    categories=["Shallow", "Intermediate", "Deep"], ordered=True
                ),
                "mag": np.dtype("float64"),
                "mag_class": pd.CategoricalDtype(
                    categories=["Light", "Moderate", "Strong", "Major"], ordered=True
                ),
                "place": pd.StringDtype("pyarrow"),
            }
        )
        pd.testing.assert_series_equal(df.dtypes, expected_dtypes)
    elif engine == "polars":
        import polars as pl

        expected_schema = {
            "time": pl.Datetime(time_unit="us", time_zone=None),
            "lat": pl.Float64,
            "lon": pl.Float64,
            "depth": pl.Float64,
            "depth_class": pl.Enum(categories=["Shallow", "Intermediate", "Deep"]),
            "mag": pl.Float64,
            "mag_class": pl.Enum(categories=["Light", "Moderate", "Strong", "Major"]),
            "place": pl.String,
        }
        assert df.collect_schema() == expected_schema
    else:
        msg = "Not valid engine"
        raise ValueError(msg)


@pytest.mark.parametrize("engine", ["pandas", "polars"])  # dask test keep failing
def test_earthquakes_category_ordering(engine):
    pytest.importorskip(engine)
    df = hvs.earthquakes(engine=engine)
    if engine == "pandas":
        import pandas as pd

        assert isinstance(df["depth_class"].dtype, pd.CategoricalDtype)
        cat_depth = df["depth_class"].cat
        assert cat_depth.ordered
        assert list(cat_depth.categories) == ["Shallow", "Intermediate", "Deep"]

        assert isinstance(df["mag_class"].dtype, pd.CategoricalDtype)
        cat_mag = df["mag_class"].cat
        assert cat_mag.ordered
        assert list(cat_mag.categories) == ["Light", "Moderate", "Strong", "Major"]
    else:
        pytest.importorskip(engine)
        import polars as pl

        schema = df.schema
        expected_depth_type_str = str(pl.Enum(["Shallow", "Intermediate", "Deep"]))
        expected_mag_type_str = str(pl.Enum(["Light", "Moderate", "Strong", "Major"]))
        assert str(schema["depth_class"]) == expected_depth_type_str
        assert str(schema["mag_class"]) == expected_mag_type_str


@pytest.mark.parametrize("engine", list(_EAGER_TABULAR_LOOKUP))
def test_apple_stocks_schema(engine):
    pytest.importorskip(engine)
    df = hvs.apple_stocks(engine=engine)
    if engine == "pandas":
        import numpy as np
        import pandas as pd

        expected_dtypes = pd.Series(
            {
                "date": np.dtype("datetime64[ns]"),
                "open": np.dtype("float64"),
                "high": np.dtype("float64"),
                "low": np.dtype("float64"),
                "close": np.dtype("float64"),
                "volume": np.dtype("int64"),
                "adj_close": np.dtype("float64"),
            }
        )
        pd.testing.assert_series_equal(df.dtypes, expected_dtypes)
    elif engine == "polars":
        import polars as pl

        assert df.schema == {
            "date": pl.Date,
            "open": pl.Float64,
            "high": pl.Float64,
            "low": pl.Float64,
            "close": pl.Float64,
            "volume": pl.Int64,
            "adj_close": pl.Float64,
        }
    else:
        msg = "Not valid engine"
        raise ValueError(msg)


@pytest.mark.parametrize("engine", list(_LAZY_TABULAR_LOOKUP))
def test_apple_stocks_schema_lazy(engine):
    pytest.importorskip(engine)
    df = hvs.apple_stocks(engine=engine, lazy=True)
    if engine == "dask":
        import numpy as np
        import pandas as pd

        expected_dtypes = pd.Series(
            {
                "date": np.dtype("datetime64[ns]"),
                "open": np.dtype("float64"),
                "high": np.dtype("float64"),
                "low": np.dtype("float64"),
                "close": np.dtype("float64"),
                "volume": np.dtype("int64"),
                "adj_close": np.dtype("float64"),
            }
        )
        pd.testing.assert_series_equal(df.dtypes, expected_dtypes)
    elif engine == "polars":
        import polars as pl

        assert df.collect_schema() == {
            "date": pl.Date,
            "open": pl.Float64,
            "high": pl.Float64,
            "low": pl.Float64,
            "close": pl.Float64,
            "volume": pl.Int64,
            "adj_close": pl.Float64,
        }
    else:
        msg = "Not valid engine"
        raise ValueError(msg)


@pytest.mark.parametrize("engine", list(_EAGER_TABULAR_LOOKUP))
def test_stocks_schema(engine):
    pytest.importorskip(engine)
    df = hvs.stocks(engine=engine)
    if engine == "pandas":
        import numpy as np
        import pandas as pd

        expected_dtypes = pd.Series(
            {
                "date": np.dtype("datetime64[ns]"),
                "Apple": np.dtype("float64"),
                "Amazon": np.dtype("float64"),
                "Google": np.dtype("float64"),
                "Meta": np.dtype("float64"),
                "Microsoft": np.dtype("float64"),
                "Netflix": np.dtype("float64"),
            }
        )
        pd.testing.assert_series_equal(df.dtypes, expected_dtypes)
    elif engine == "polars":
        import polars as pl

        assert df.schema == {
            "date": pl.Date,
            "Apple": pl.Float64,
            "Amazon": pl.Float64,
            "Google": pl.Float64,
            "Meta": pl.Float64,
            "Microsoft": pl.Float64,
            "Netflix": pl.Float64,
        }
    else:
        msg = "Not valid engine"
        raise ValueError(msg)


@pytest.mark.parametrize("engine", list(_LAZY_TABULAR_LOOKUP))
def test_stocks_schema_lazy(engine):
    pytest.importorskip(engine)
    df = hvs.stocks(engine=engine, lazy=True)
    if engine == "dask":
        import numpy as np
        import pandas as pd

        expected_dtypes = pd.Series(
            {
                "date": np.dtype("datetime64[ns]"),
                "Apple": np.dtype("float64"),
                "Amazon": np.dtype("float64"),
                "Google": np.dtype("float64"),
                "Meta": np.dtype("float64"),
                "Microsoft": np.dtype("float64"),
                "Netflix": np.dtype("float64"),
            }
        )
        pd.testing.assert_series_equal(df.dtypes, expected_dtypes)
    elif engine == "polars":
        import polars as pl

        assert df.collect_schema() == {
            "date": pl.Date,
            "Apple": pl.Float64,
            "Amazon": pl.Float64,
            "Google": pl.Float64,
            "Meta": pl.Float64,
            "Microsoft": pl.Float64,
            "Netflix": pl.Float64,
        }
    else:
        msg = "Not valid engine"
        raise ValueError(msg)


@pytest.mark.parametrize("engine", list(_EAGER_TABULAR_LOOKUP))
def test_synthetic_clusters_schema(engine):
    pytest.importorskip(engine)
    df = hvs.synthetic_clusters(engine=engine)
    cats = ["d1", "d2", "d3", "d4", "d5"]
    if engine == "pandas":
        import numpy as np
        import pandas as pd

        expected_dtypes = pd.Series(
            {
                "x": np.dtype("float64"),
                "y": np.dtype("float64"),
                "s": np.dtype("float64"),
                "val": np.dtype("int64"),
                "cat": pd.CategoricalDtype(categories=cats, ordered=False),
            }
        )
        pd.testing.assert_series_equal(df.dtypes, expected_dtypes)
    elif engine == "polars":
        import polars as pl

        assert df.schema == {
            "x": pl.Float64,
            "y": pl.Float64,
            "s": pl.Float64,
            "val": pl.Int64,
            "cat": pl.Enum(categories=cats),
        }
    else:
        msg = "Not valid engine"
        raise ValueError(msg)


@pytest.mark.parametrize("engine", list(_LAZY_TABULAR_LOOKUP))
def test_synthetic_clusters_schema_lazy(engine):
    pytest.importorskip(engine)
    df = hvs.synthetic_clusters(engine=engine, lazy=True)
    cats = ["d1", "d2", "d3", "d4", "d5"]
    if engine == "dask":
        import numpy as np
        import pandas as pd

        expected_dtypes = pd.Series(
            {
                "x": np.dtype("float64"),
                "y": np.dtype("float64"),
                "s": np.dtype("float64"),
                "val": np.dtype("int64"),
                "cat": pd.CategoricalDtype(categories=cats, ordered=False),
            }
        )
        pd.testing.assert_series_equal(df.dtypes, expected_dtypes)
    elif engine == "polars":
        import polars as pl

        assert df.collect_schema() == {
            "x": pl.Float64,
            "y": pl.Float64,
            "s": pl.Float64,
            "val": pl.Int64,
            "cat": pl.Enum(categories=cats),
        }
    else:
        msg = "Not valid engine"
        raise ValueError(msg)


@pytest.mark.parametrize("engine", list(_EAGER_TABULAR_LOOKUP))
def test_synthetic_clusters_total_points(engine):
    pytest.importorskip(engine)
    df = hvs.synthetic_clusters(engine=engine, total_points=10)
    assert len(df) == 10
    with pytest.raises(ValueError, match="total_points must be a multiple of 5"):
        hvs.synthetic_clusters(engine=engine, total_points=11)


@pytest.mark.parametrize("engine", list(_LAZY_TABULAR_LOOKUP))
def test_synthetic_clusters_lazy_total_points(engine):
    pytest.importorskip(engine)
    df = hvs.synthetic_clusters(engine=engine, lazy=True, total_points=10)
    if engine == "polars":
        df = df.collect()
    assert len(df) == 10
    with pytest.raises(ValueError, match="total_points must be a multiple of 5"):
        hvs.synthetic_clusters(engine=engine, lazy=True, total_points=11)


@pytest.mark.parametrize("engine", ["geopandas"])
def test_us_states_schema(engine):
    pytest.importorskip(engine)
    df = hvs.us_states(engine=engine)

    if engine == "geopandas":
        import numpy as np
        import pandas as pd

        expected_dtypes = pd.Series(
            {
                "state": np.dtype("O"),
                "median_income": np.dtype("float64"),
                "income_range": pd.CategoricalDtype(),
                "pop_density": np.dtype("float64"),
                "pop_density_range": pd.CategoricalDtype(),
                "bea_region": pd.CategoricalDtype(),
                "geometry": "geometry",
            }
        )

        actual_dtypes = df.dtypes.astype(str)
        for col, expected in expected_dtypes.items():
            assert col in df.columns
            if expected == "geometry":
                assert actual_dtypes[col].startswith("geometry")
            else:
                assert actual_dtypes[col] == str(expected)
    else:
        msg = "Not valid engine"
        raise ValueError(msg)


@pytest.mark.parametrize("engine", ["geopandas"])
def test_us_states_category_ordering(engine):
    pytest.importorskip(engine)
    df = hvs.us_states(engine=engine)
    if engine == "geopandas":
        import pandas as pd

        assert isinstance(df["income_range"].dtype, pd.CategoricalDtype)
        cat_income_range = df["income_range"].cat
        assert cat_income_range.ordered
        assert list(cat_income_range.categories) == [
            "<$40k",
            "$40k-$50k",
            "$50k-$60k",
            "$60k-$70k",
            ">$70k",
        ]

        assert isinstance(df["pop_density_range"].dtype, pd.CategoricalDtype)
        cat_pop_density_range = df["pop_density_range"].cat
        assert cat_pop_density_range.ordered
        assert list(cat_pop_density_range.categories) == [
            "Very Low",
            "Low",
            "Moderate",
            "High",
            "Very High",
        ]
    else:
        msg = "Not valid engine"
        raise ValueError(msg)


@pytest.mark.parametrize("engine", list(_EAGER_TABULAR_LOOKUP))
def test_nyc_taxi_schema_eager(engine):
    pytest.importorskip(engine)
    df = hvs.nyc_taxi_remote(engine=engine)
    if engine == "pandas":
        import numpy as np
        import pandas as pd

        expected_dtypes = pd.Series(
            {
                "tpep_pickup_datetime": np.dtype("datetime64[us]"),
                "tpep_dropoff_datetime": np.dtype("datetime64[us]"),
                "passenger_count": np.dtype("uint8"),
                "trip_distance": np.dtype("float32"),
                "pickup_x": np.dtype("float32"),
                "pickup_y": np.dtype("float32"),
                "dropoff_x": np.dtype("float32"),
                "dropoff_y": np.dtype("float32"),
                "fare_amount": np.dtype("float32"),
                "tip_amount": np.dtype("float32"),
                "dropoff_hour": np.dtype("uint8"),
                "pickup_hour": np.dtype("uint8"),
            }
        )
        pd.testing.assert_series_equal(df.dtypes, expected_dtypes)
    elif engine == "polars":
        import polars as pl

        expected_schema = {
            "tpep_pickup_datetime": pl.Datetime(time_unit="ns", time_zone=None),
            "tpep_dropoff_datetime": pl.Datetime(time_unit="ns", time_zone=None),
            "passenger_count": pl.UInt8,
            "trip_distance": pl.Float32,
            "pickup_x": pl.Float32,
            "pickup_y": pl.Float32,
            "dropoff_x": pl.Float32,
            "dropoff_y": pl.Float32,
            "fare_amount": pl.Float32,
            "tip_amount": pl.Float32,
            "dropoff_hour": pl.UInt8,
            "pickup_hour": pl.UInt8,
        }
        assert df.schema == expected_schema
    else:
        msg = "Not valid engine"
        raise ValueError(msg)


@pytest.mark.parametrize("engine", list(_LAZY_TABULAR_LOOKUP))
def test_nyc_taxi_schema_lazy(engine):
    pytest.importorskip(engine)
    df = hvs.nyc_taxi_remote(engine=engine, lazy=True)
    if engine == "dask":
        import numpy as np
        import pandas as pd

        expected_dtypes = pd.Series(
            {
                "tpep_pickup_datetime": np.dtype("datetime64[us]"),
                "tpep_dropoff_datetime": np.dtype("datetime64[us]"),
                "passenger_count": np.dtype("uint8"),
                "trip_distance": np.dtype("float32"),
                "pickup_x": np.dtype("float32"),
                "pickup_y": np.dtype("float32"),
                "dropoff_x": np.dtype("float32"),
                "dropoff_y": np.dtype("float32"),
                "fare_amount": np.dtype("float32"),
                "tip_amount": np.dtype("float32"),
                "dropoff_hour": np.dtype("uint8"),
                "pickup_hour": np.dtype("uint8"),
            }
        )
        pd.testing.assert_series_equal(df.dtypes, expected_dtypes)
    elif engine == "polars":
        import polars as pl

        expected_schema = {
            "tpep_pickup_datetime": pl.Datetime(time_unit="ns", time_zone=None),
            "tpep_dropoff_datetime": pl.Datetime(time_unit="ns", time_zone=None),
            "passenger_count": pl.UInt8,
            "trip_distance": pl.Float32,
            "pickup_x": pl.Float32,
            "pickup_y": pl.Float32,
            "dropoff_x": pl.Float32,
            "dropoff_y": pl.Float32,
            "fare_amount": pl.Float32,
            "tip_amount": pl.Float32,
            "dropoff_hour": pl.UInt8,
            "pickup_hour": pl.UInt8,
        }
        assert df.lazy().collect().schema == expected_schema
    else:
        msg = "Not valid engine"
        raise ValueError(msg)


@pytest.mark.parametrize("engine", list(_EAGER_TABULAR_LOOKUP))
def test_nyc_taxi_colums_eager(engine):
    pytest.importorskip(engine)

    kwargs = {"columns": ["pickup_x", "pickup_y", "dropoff_x", "dropoff_y"]}
    df = hvs.nyc_taxi_remote(engine=engine, lazy=False, engine_kwargs=kwargs)

    expected_columns = {"pickup_x", "pickup_y", "dropoff_x", "dropoff_y"}
    if engine in ("pandas", "polars"):
        assert set(df.columns) == expected_columns
        assert len(df.columns) == 4
    else:
        msg = "Not valid engine"
        raise ValueError(msg)


@pytest.mark.parametrize("engine", ["dask"])
def test_nyc_taxi_columns_lazy(engine):
    # Test only dask for lazy loading as polars `scan_parquet` has no `columns` parameter
    pytest.importorskip(engine)

    kwargs = {"columns": ["pickup_x", "pickup_y", "dropoff_x", "dropoff_y"]}
    expected_columns = {"pickup_x", "pickup_y", "dropoff_x", "dropoff_y"}

    df = hvs.nyc_taxi_remote(engine=engine, engine_kwargs=kwargs, lazy=True)
    columns = df.columns
    assert set(columns) == expected_columns
    assert len(columns) == 4



"""hvsampledata: shared datasets for the HoloViz projects.

Currently available datasets:

| Name               | Type    | Included |
| ------------------ | ------- | -------- |
| air_temperature    | Gridded | Yes      |
| apple_stocks       | Tabular | Yes      |
| earthquakes        | Tabular | Yes      |
| landsat_rgb        | Gridded | Yes      |
| nyc_taxi_remote    | Tabular | Remote   |
| penguins           | Tabular | Yes      |
| penguins_rgba      | Gridded | Yes      |
| stocks             | Tabular | Yes      |
| synthetic_clusters | Tabular | Yes      |
| us_states          | Tabular | Yes      |

Use it with:

>>> import hvsampledata
>>> ds = hvsampledata.air_temperature("xarray")
>>> df = hvsampledata.penguins("pandas")

"""

from __future__ import annotations

from typing import Any

from .__version import __version__
from ._util import _DATAPATH, _load_gridded, _load_tabular

# -----------------------------------------------------------------------------
# Tabular data
# -----------------------------------------------------------------------------


def synthetic_clusters(
    engine: str,
    *,
    lazy: bool = False,
    total_points: int = 1_000_000,
):
    """Large tabular dataset with 5 synthetic clusters generated from a normal
    distribution with a scale distributed roughly according to a power law.

    Parameters
    ----------
    engine : str
        Engine used to read the dataset. "pandas" or "polars" for eager dataframes,
        "polars" or "dask" for lazy dataframes (polars need lazy=True).
    lazy : bool, optional
        Whether to load the dataset in a lazy container, by default False.
    total_points: int, default=1_000_000
        Total number of points in the dataset returned, must be a multiple of 5.

    Description
    -----------
    Synthetic dataset that contains 5 clusters, each cluster generated from a
    normal distribution centered on the first two values of these tuples (e.g.
    `x,y=2,2` for the first cluster) and with a standard deviation equal to
    the third value (e.g. `s=0.03` for the first cluster):

    ```python
    clusters = [
        (2, 2, 0.03, 0, "d1"),
        (2, -2, 0.10, 1, "d2"),
        (-2, -2, 0.50, 2, "d3"),
        (-2, 2, 1.00, 3, "d4"),
        (0, 0, 3.00, 4, "d5"),
    ]
    ```

    The standard deviation / scale is distributed roughly according to a power
    law, with `s ~= 0.005 * x^4`.

    Schema
    ------
    | name | type        | description                                              |
    |:-----|:------------|:---------------------------------------------------------|
    | x    | number      | x coordinate                                             |
    | y    | number      | y coordinate                                             |
    | s    | number      | standard deviation of the distribution                   |
    | val  | integer     | integer value per distribution, one of 0, 1, 2, 3, 4     |
    | cat  | categorical | string value per distribution, one of d1, d2, d3, d4, d5 |
    """

    clusters = [
        (2, 2, 0.03, 0, "d1"),
        (2, -2, 0.10, 1, "d2"),
        (-2, -2, 0.50, 2, "d3"),
        (-2, 2, 1.00, 3, "d4"),
        (0, 0, 3.00, 4, "d5"),
    ]

    if total_points % 5:
        msg = "total_points must be a multiple of 5"
        raise ValueError(msg)
    points_per_cluster = total_points // 5
    cats = ["d1", "d2", "d3", "d4", "d5"]
    if engine in ["pandas", "dask"]:
        import numpy as np
        import pandas as pd

        cat_dtype = pd.CategoricalDtype(categories=cats, ordered=False)

        def create_synthetic_dataset(x, y, s, val, cat, cat_dtype, num, dask=False):
            seed = np.random.default_rng(1)
            df = pd.DataFrame(
                {
                    "x": seed.normal(x, s, num),
                    "y": seed.normal(y, s, num),
                    "s": s,
                    "val": val,
                    "cat": pd.Series([cat] * num, dtype=cat_dtype),
                }
            )
            if dask:
                import dask.dataframe as dd

                return dd.from_pandas(df, npartitions=2)
            return df

        if engine == "pandas":
            func_concat = pd.concat
            kwargs_concat = {"ignore_index": True}
        elif engine == "dask":
            import dask.dataframe as dd

            func_concat = dd.concat
            kwargs_concat = {"axis": 0}  # , "interleave_partitions":  True}
        df = func_concat(
            [
                create_synthetic_dataset(
                    x, y, s, val, cat, cat_dtype, points_per_cluster, dask=engine == "dask"
                )
                for x, y, s, val, cat in clusters
            ],
            **kwargs_concat,
        )
        return df
    elif engine == "polars":
        import random

        import polars as pl

        def create_synthetic_dataset(x, y, s, val, cat, num, lazy=False):
            pdf = pl.DataFrame(
                {
                    "x": [random.gauss(x, s) for _ in range(num)],
                    "y": [random.gauss(y, s) for _ in range(num)],
                    "s": [s] * num,
                    "val": [val] * num,
                    "cat": pl.Series([cat] * num).cast(pl.Enum(cats)),
                }
            )
            if lazy:
                return pdf.lazy()
            return pdf

        # Use a global StringCache so categoricals are shared
        with pl.StringCache():
            dfp = pl.concat(
                [
                    create_synthetic_dataset(x, y, s, val, cat, points_per_cluster, lazy=lazy)
                    for x, y, s, val, cat in clusters
                ],
                how="vertical",
            )
        return dfp


def penguins(
    engine: str,
    *,
    engine_kwargs: dict[str, Any] | None = None,
    lazy: bool = False,
):
    """Penguins tabular dataset.

    Parameters
    ----------
    engine : str
        Engine used to read the dataset. "pandas" or "polars" for eager dataframes,
        "polars" or "dask" for lazy dataframes (lazy=True).
    engine_kwargs : dict[str, Any], optional
        Additional kwargs to pass to `read_csv`, by default None.
    lazy : bool, optional
        Whether to load the dataset in a lazy container, by default False.

    Description
    -----------
    Tabular records of morphological measurements and demographic information
    from 344 penguins. There are 3 different species of penguins in this dataset,
    collected from 3 islands in the Palmer Archipelago, Antarctica.

    Data were collected and made available by Dr. Kristen Gorman and the Palmer
    Station, Antarctica LTER, a member of the Long Term Ecological Research Network.

    Schema
    ------
    | name              | type    | description                                                         |
    |:------------------|:--------|:--------------------------------------------------------------------|
    | species           | string  | Penguin species (Adelie, Gentoo, or Chinstrap)                      |
    | island            | string  | Island where the penguin was observed (Torgersen, Biscoe, or Dream) |
    | bill_length_mm    | number  | Bill/Beak length in millimeter                                      |
    | bill_depth_mm     | number  | Bill/Beak depth in millimeters                                      |
    | flipper_length_mm | number* | Flipper length in millimeters                                       |
    | body_mass_g       | number* | Body mass in grams                                                  |
    | sex               | string  | Sex of the penguin (male, female or null)                           |
    | year              | integer | Observation year                                                    |

    * float64 for pandas and dask, int64 for polars

    Source
    ------
    `penguins.csv` dataset from the R `palmerpenguins` package
    https://github.com/allisonhorst/palmerpenguins.

    License
    -------
    Data are available by CC-0 license in accordance with the Palmer Station LTER
    Data Policy and the LTER Data Access Policy for Type I data.

    References
    ----------
    Data originally published in:

    Gorman KB, Williams TD, Fraser WR (2014). Ecological sexual dimorphism and
    environmental variability within a community of Antarctic penguins (genus
    Pygoscelis). PLoS ONE 9(3):e90081. https://doi.org/10.1371/journal.pone.0090081
    """
    if engine == "polars":
        engine_kwargs = {"null_values": "NA"} | (engine_kwargs or {})
    tab = _load_tabular(
        "penguins.csv",
        format="csv",
        engine=engine,
        engine_kwargs=engine_kwargs,
        lazy=lazy,
    )
    return tab


def earthquakes(
    engine: str,
    *,
    engine_kwargs: dict[str, Any] | None = None,
    lazy: bool = False,
):
    """Earthquakes tabular dataset.

    Parameters
    ----------
    engine : str
        Engine used to read the dataset. "pandas" or "polars" for eager dataframes,
        "polars" or "dask" for lazy dataframes (lazy=True).
    engine_kwargs : dict[str, Any], optional
        Additional kwargs to pass to `read_csv`, by default None.
    lazy : bool, optional
        Whether to load the dataset in a lazy container, by default False.

    Description
    -----------
    Tabular record of earthquake events from the USGS Earthquake Catalog that provides detailed
    information including parameters such as time, location as latitude/longitude coordinates
    and place name, depth, and magnitude. The dataset contains 596 events.

    Note: The columns `depth_class` and `mag_class` were created by categorizing numerical values from
    the `depth` and `mag` columns in the original dataset using custom-defined binning:

    Depth Classification

    | depth     | depth_class  |
    |-----------|--------------|
    | Below 70  | Shallow      |
    | 70 - 300  | Intermediate |
    | Above 300 | Deep         |

    Magnitude Classification

    | mag         | mag_class |
    |-------------|-----------|
    | 3.9 - <4.9  | Light     |
    | 4.9 - <5.9  | Moderate  |
    | 5.9 - <6.9  | Strong    |
    | 6.9 - <7.9  | Major     |


    Schema
    ------
    | name        | type       | description                                                         |
    |:------------|:-----------|:--------------------------------------------------------------------|
    | time        | datetime   | UTC Time when the event occurred.                                   |
    | lat         | float      | Decimal degrees latitude. Negative values for southern latitudes.   |
    | lon         | float      | Decimal degrees longitude. Negative values for western longitudes   |
    | depth       | float      | Depth of the event in kilometers.                                   |
    | depth_class | category   | The depth category derived from the depth column.                   |
    | mag         | float      | The magnitude for the event.                                        |
    | mag_class   | category   | The magnitude category derived from the mag column.                 |
    | place       | string     | Textual description of named geographic region near to the event.   |

    Source
    ------
    `earthquakes.csv` dataset courtesy of the U.S. Geological Survey
    https://www.usgs.gov/programs/earthquake-hazards, with 4 months of data selected
    from April to July 2024 along the Pacific Ring of Fire region (lat=(-10,10), lon=(110,140))

    License
    -------
    U.S. Public domain
    Data available from U.S. Geological Survey, National Geospatial Program.
    Visit the USGS at https://usgs.gov.

    """
    depth_order = ["Shallow", "Intermediate", "Deep"]
    mag_order = ["Light", "Moderate", "Strong", "Major"]
    engine_kwargs = engine_kwargs or {}

    # convert `time` column to datetime and `mag_class` and `depth_class` to categories
    if engine == "polars":
        import polars as pl

        engine_kwargs = {
            "try_parse_dates": True,
            "schema_overrides": {
                "depth_class": pl.Enum(depth_order),
                "mag_class": pl.Enum(mag_order),
            },
        } | engine_kwargs
    else:
        import pandas as pd

        engine_kwargs = {
            "parse_dates": ["time"],
            "dtype": {
                "depth_class": pd.api.types.CategoricalDtype(categories=depth_order, ordered=True),
                "mag_class": pd.api.types.CategoricalDtype(categories=mag_order, ordered=True),
            },
        } | engine_kwargs
    return _load_tabular(
        "earthquakes.csv",
        format="csv",
        engine=engine,
        engine_kwargs=engine_kwargs,
        lazy=lazy,
    )


def apple_stocks(
    engine: str,
    *,
    engine_kwargs: dict[str, Any] | None = None,
    lazy: bool = False,
):
    """Apple Inc. (AAPL) stocks dataset.

    Parameters
    ----------
    engine : str
        Engine used to read the dataset. "pandas" or "polars" for eager dataframes,
        "polars" or "dask" for lazy dataframes (lazy=True).
    engine_kwargs : dict[str, Any], optional
        Additional kwargs to pass to `read_csv`, by default None.
    lazy : bool, optional
        Whether to load the dataset in a lazy container, by default False.

    Description
    -----------
    Tabular record of Apple Inc. (AAPL) daily stock trading data from the U.S. stock market
    from January 2019 to December 2024.
    Each row represents a single trading day with pricing and volume information.

    This dataset contains 1509 rows and was collected from public news sources.

    Schema
    ------
    | name       | type     | description                                               |
    |:-----------|:-------- |:----------------------------------------------------------|
    | date       | datetime | The trading date                                          |
    | open       | float    | Opening price of the stock on that day                    |
    | high       | float    | Highest price of the stock during the trading day         |
    | low        | float    | Lowest price of the stock during the trading day          |
    | close      | float    | Closing price of the stock on that day                    |
    | volume     | integer  | Number of shares traded                                   |
    | adj_close  | float    | Adjusted closing price reflecting splits and dividends    |

    Source
    ------
    `apple_stocks.csv` dataset generated from historical data for Apple Inc. (AAPL) sourced from Yahoo Finance.

    License
    -------
    Data provided for demonstration and educational purposes only.
    Users must ensure compliance with the original data providers terms of use.
    See https://legal.yahoo.com/us/en/yahoo/terms/product-atos/apiforydn/index.html
    """
    engine_kwargs = engine_kwargs or {}
    # convert `date` column to datetime object
    if engine == "polars":
        engine_kwargs = {
            "try_parse_dates": True,
        } | engine_kwargs
    else:
        engine_kwargs = {
            "parse_dates": ["date"],
        } | engine_kwargs
    return _load_tabular(
        "apple_stocks.csv",
        format="csv",
        engine=engine,
        engine_kwargs=engine_kwargs,
        lazy=lazy,
    )


def stocks(
    engine: str,
    *,
    engine_kwargs: dict[str, Any] | None = None,
    lazy: bool = False,
):
    """Selected stocks dataset.

    Parameters
    ----------
    engine : str
        Engine used to read the dataset. "pandas" or "polars" for eager dataframes,
        "polars" or "dask" for lazy dataframes (lazy=True).
    engine_kwargs : dict[str, Any], optional
        Additional kwargs to pass to `read_csv`, by default None.
    lazy : bool, optional
        Whether to load the dataset in a lazy container, by default False.

    Description
    -----------
    Tabular dataset containing weekly rebased stock prices for selected Tech companies:
    Apple, Amazon, Google, Meta, Microsoft, and Netflix from January 2019 to December 2023.
    The stock prices have been rebased to start at 1.0 from the first row.

    This dataset contains 261 rows and can be used to compare the relative performance of
    each company's stock over that time period.

    Schema
    ------
    | name       | type     | description                            |
    |:-----------|:---------|:---------------------------------------|
    | date       | datetime | The trading date (weekly interval)     |
    | Apple      | float    | Normalized price of Google stock       |
    | Amazon     | float    | Normalized price of Apple stock        |
    | Google     | float    | Normalized price of Amazon stock       |
    | Meta       | float    | Normalized price of Facebook stock     |
    | Microsoft  | float    | Normalized price of Netflix stock      |
    | Netflix    | float    | Normalized price of Microsoft stock    |

    Source
    ------
    `stocks.csv` dataset derived from historical stock prices of selected Tech companies,
    sourced from Yahoo Finance and rebased for comparative analysis.

    License
    -------
    Data provided for educational and demonstration purposes only.
    Users must ensure compliance with the original data providers terms of use.
    See https://legal.yahoo.com/us/en/yahoo/terms/product-atos/apiforydn/index.html
    """
    engine_kwargs = engine_kwargs or {}
    # convert `date` column to datetime object
    if engine == "polars":
        engine_kwargs = {
            "try_parse_dates": True,
        } | engine_kwargs
    else:
        engine_kwargs = {
            "parse_dates": ["date"],
        } | engine_kwargs
    return _load_tabular(
        "stocks.csv",
        format="csv",
        engine=engine,
        engine_kwargs=engine_kwargs,
        lazy=lazy,
    )


def us_states(
    engine: str,
    *,
    engine_kwargs: dict[str, Any] | None = None,
):
    """U.S. States socio-economic and geographic dataset.

    Parameters
    ----------
    engine : str
        Engine used to read the dataset. Only `"geopandas"` is supported.
    engine_kwargs : dict[str, Any], optional
        Additional kwargs to pass to `geopandas.read_file`, by default None.

    Description
    -----------
    Geodataframe with demographic and economic data for U.S. states, including:
    - median income
    - population density
    - BEA-defined economic region
    - classified income and population density ranges

    The dataset contains 49 rows for the 50 US States, minus Hawaii that
    was not included as it makes geographic plotting less user-friendly.

    The polygons and multipolygons are in the WGS84 coordinate reference
    system (EPSG:4326).

    Schema
    ------
    | name              | type      | description                                  |
    |:------------------|:----------|:---------------------------------------------|
    | state             | string    | U.S. state name                              |
    | median_income     | float     | Median household income                      |
    | income_range      | category  | Binned income range                          |
    | pop_density       | float     | Population density per square mile           |
    | pop_density_range | category  | Binned population density                    |
    | bea_region        | category  | U.S. economic region from the BEA            |
    | geometry          | geometry  | Polygon/MultiPolygon geometry for each state |

    Source
    ------
    Custom dataset derived from U.S. Census and BEA data.

    License
    -------
    Public domain / derived from U.S. government data.
    """
    if engine != "geopandas":
        msg = "us_states dataset only supports 'geopandas' engine"
        raise ValueError(msg)

    import geopandas as gpd
    import pandas as pd

    fp = _DATAPATH / "us_states.geojson"
    engine_kwargs = {} if engine_kwargs is None else engine_kwargs
    gdf = gpd.read_file(fp, **engine_kwargs)

    income_cats = ["<$40k", "$40k-$50k", "$50k-$60k", "$60k-$70k", ">$70k"]
    pop_density_cats = ["Very Low", "Low", "Moderate", "High", "Very High"]
    gdf["income_range"] = pd.Categorical(gdf["income_range"], categories=income_cats, ordered=True)
    gdf["pop_density_range"] = pd.Categorical(
        gdf["pop_density_range"], categories=pop_density_cats, ordered=True
    )
    gdf["bea_region"] = gdf["bea_region"].astype("category")
    return gdf


def nyc_taxi_remote(
    engine: str,
    *,
    engine_kwargs: dict[str, Any] | None = None,
    lazy: bool = False,
):
    """NYC Taxi trip record data 2015.

    Parameters
    ----------
    engine : str
        Engine used to read the dataset. "pandas" or "polars" for eager dataframes,
        "polars" or "dask" for lazy dataframes.
    engine_kwargs : dict[str, Any], optional
        Additional kwargs to pass to `read_parquet`/`scan_parquet`, by default None.
        Note: For polars lazy loading, column selection is applied via .select() after scan_parquet().
    lazy : bool, optional
        Whether to load the dataset in a lazy container.

    Description
    -----------
    Tabular record of New York City taxi trip data from 2015, containing detailed information
    about yellow taxi trips including pickup and dropoff locations, trip characteristics,
    and fare details. Each row represents a single taxi trip with geographic coordinates,
    timestamps, and payment information.

    This dataset contains over 11 million taxi trip records and is approximately 260MB in size.
    The data has been pre-processed and optimized for efficient storage and faster loading.
    Coordinates have been transformed to Web Mercator projection.

    The dataset is downloaded directly from the HoloViz S3 bucket as a parquet file here:
    https://datasets.holoviz.org/nyc_taxi/v2/nyc_taxi_wide.parq.

    Schema
    ------
    | name                  | type      | description                              |
    |:----------------------|:----------|:-----------------------------------------|
    | tpep_pickup_datetime  | datetime  | Trip pickup timestamp (US/Eastern time)  |
    | tpep_dropoff_datetime | datetime  | Trip dropoff timestamp (US/Eastern time) |
    | passenger_count       | uint8     | Number of passengers                     |
    | trip_distance         | float32   | Trip distance in miles                   |
    | pickup_x              | float32   | Pickup X coordinate                      |
    | pickup_y              | float32   | Pickup Y coordinate                      |
    | dropoff_x             | float32   | Dropoff X coordinate                     |
    | dropoff_y             | float32   | Dropoff Y coordinate                     |
    | fare_amount           | float32   | Base fare in dollars                     |
    | tip_amount            | float32   | Tip amount in dollars                    |
    | dropoff_hour          | uint8     | Dropoff time in 24hr format              |
    | pickup_hour           | uint8     | Pickup time in 24hr format               |

    Source
    ------
    NYC Taxi and Limousine Commission (TLC) trip record data.
    Available at: https://www.nyc.gov/site/tlc/about/tlc-trip-record-data.page

    License
    -------
    Public domain / NYC Open Data.

    Examples
    --------
    Load full dataset:

    >>> df = nyc_taxi_remote("polars")  # Polars DataFrame
    >>> df = nyc_taxi_remote("dask", lazy=True)  # Dask DataFrame

    Load specific columns:

    >>> df = nyc_taxi_remote("polars", engine_kwargs={"columns": ["pickup_x", "pickup_y"]})
    """
    engine_kwargs = engine_kwargs or {}

    data = _load_tabular(
        "https://datasets.holoviz.org/nyc_taxi/v2/nyc_taxi_wide.parq",
        format="parquet",
        engine=engine,
        engine_kwargs=engine_kwargs,
        lazy=lazy,
    )

    return data


# -----------------------------------------------------------------------------
# Gridded data
# -----------------------------------------------------------------------------


def air_temperature(
    engine: str,
    *,
    engine_kwargs=None,
):
    """Air Temperature gridded dataset.

    Parameters
    ----------
    engine : str
        Engine used to read the dataset, "xarray" is the only option available.
    engine_kwargs : dict[str, Any], optional
        Additional kwargs to pass to `xarray.open_dataset`, by default None.

    Description
    -----------
    The NCEP/NCAR Reanalysis 1 project is using a state-of-the-art analysis/forecast
    system to perform data assimilation using past data from 1948 to the present.

    This dataset was created by temporally resampling the `air_temperature` dataset
    made available by xarray-data, itself being a spatial and temporal subset of
    the original data. It only includes the air temperature variable.

    Temporal coverage:
    - Every 6 hours, starting from 00:00
    - 2014-02-24 to 2014-02-28

    Spatial coverage:
    - 2.5 degree x 2.5 degree grid (lon:53xlat:25)
    - ~ North America (lon:[200-330], lat:[15-75])

    Dimensions:
    - lat: float32, 25 values
    - lon: float32, 53 values
    - time: datetime64[ns], 20 values

    Variables:
    - air: [time|lat|lon], float64, air temperature in Kelvin

    Source
    ------
    `air_temperature.nc` dataset from the `xarray-data` Github repository
    https://github.com/pydata/xarray-data, resampled to 20 timestamps between
    2014-02-24 and 2014-02-28.

    Original data from:
    https://psl.noaa.gov/data/gridded/data.ncep.reanalysis.html

    License
    -------
    NCEP-NCAR Reanalysis 1 data provided by the NOAA PSL, Boulder, Colorado, USA,
    from their website at https://psl.noaa.gov

    References
    ----------
    Kalnay et al.,The NCEP/NCAR 40-year reanalysis project, Bull. Amer. Meteor. Soc., 77, 437-470, 1996
    """
    ds = _load_gridded(
        "air_temperature_small.nc",
        format="dataset",
        engine=engine,
        engine_kwargs=engine_kwargs,
    )
    if str(ds.dtypes["air"]) == "float32":
        # Float32 with older version of xarray/netcdf4.
        ds = ds.astype("float64")
    return ds


def penguins_rgba(
    engine: str,
):
    """Penguins RGBA image.

    Parameters
    ----------
    engine : str
        Engine used to read the dataset, only `xarray` is available.

    Description
    -----------
    This dataset is an image of two adult Emperor penguins with a juvenile in
    Antarctica, read from a 100x100 PNG file into an xarray Dataset object.

    Dimensions:
    - y: int64, 100 values
    - x: int64, 100 values
    - channel: U1, 4 values

    Variables:
    - rgba: [y|x|channel]: uint8

    Source
    ------
    Original PNG file obtained from
    https://en.wikipedia.org/wiki/Emperor_penguin#/media/File:Aptenodytes_forsteri_-Snow_Hill_Island,_Antarctica_-adults_and_juvenile-8.jpg
    Original picture from Ian Duffy from UK - Animal Portraits.
    The original file was reduced to 100x100.

    License
    -------
    Creative Commons Attribution 2.0 Generic
    https://creativecommons.org/licenses/by/2.0/deed.en
    """

    if engine != "xarray":
        msg = "xarray is the only supported engine"
        raise ValueError(msg)

    import numpy as np
    import xarray as xr
    from PIL import Image

    with Image.open(_DATAPATH / "penguins.png") as img:
        img_array = np.array(img)

    return xr.DataArray(
        img_array,
        dims=["y", "x", "channel"],
        coords={
            "y": np.arange(img_array.shape[0]),
            "x": np.arange(img_array.shape[1]),
            "channel": ["R", "G", "B", "A"],
        },
        name="rgba",
    ).to_dataset()


def landsat_rgb(
    engine: str,
    *,
    engine_kwargs=None,
):
    """Landsat 7 RGB tile.

    Parameters
    ----------
    engine : str
        Engine used to read the dataset, only `rioxarray` is available.
    engine_kwargs : dict[str, Any], optional
        Additional kwargs to pass to `rioxarray.open_rasterio`, by default None.

    Description
    -----------
    This dataset is a satellite image from the USGS Landsat 7 mission, saved
    as RGB GeoTIFF file. The area covers part of the Bahama, with pixel size of
    ~1 km and a grid of 237x215. The nodata value is 0. Its CRS is WGS 84 / UTM
    zone 18N (EPSG:32618).

    Dimensions:
    - band: int64, 3 values
    - x: float64, 237 values
    - y: float64, 215 values

    Extra coords:
    - spatial_ref: int64

    Variables:
    - rgb: [band|y|x]: uint8

    Source
    ------
    Original GeoTIFF file obtained from the `rioxarray` Github repository
    https://github.com/rasterio/rasterio/blob/95f1f5fb55be6763bf987f42eb54c644604c6d3d/tests/data/RGB.byte.tif

    The original file was downsampled with this GDAL command to make it smaller:

        gdal_translate -outsize 30% 30% RGB.byte.tif landsat_rgb.tif

    Original data from the USGS Landsat satellite 7 mission.

    License
    -------
    rasterio image RGB.byte.tif is licensed under the CC0 1.0 Universal (CC0 1.0)
    Public Domain Dedication: http://creativecommons.org/publicdomain/zero/1.0/.
    """
    if engine != "rioxarray":
        msg = "rioxarray is the only supported engine"
        raise ValueError(msg)

    import rioxarray

    fp = _DATAPATH / "landsat_rgb.tif"
    engine_kwargs = {} if engine_kwargs is None else engine_kwargs

    return rioxarray.open_rasterio(fp, **engine_kwargs).to_dataset(name="rgb")


__all__ = (
    "__version__",
    "air_temperature",
    "apple_stocks",
    "earthquakes",
    "landsat_rgb",
    "nyc_taxi_remote",
    "penguins",
    "penguins_rgba",
    "stocks",
    "synthetic_clusters",
    "us_states",
)



# !!! THIS FILE IS AUTOGENERATED !!!
# For changes see scripts/generate_typehints.py

from __future__ import annotations

from typing import Any, Literal, overload

import dask.dataframe as _dd
import geopandas as _gpd
import pandas as _pd
import polars as _pl
import xarray as _xr

__version__: str
__all__: tuple[str, ...]

@overload
def penguins(
    engine: Literal["pandas"],
    *,
    engine_kwargs: dict[str, Any] | None = None,
    lazy: Literal[False] = False,
) -> _pd.DataFrame: ...
@overload
def penguins(
    engine: Literal["polars"],
    *,
    engine_kwargs: dict[str, Any] | None = None,
    lazy: Literal[False] = False,
) -> _pl.DataFrame: ...
@overload
def penguins(
    engine: Literal["polars"],
    *,
    engine_kwargs: dict[str, Any] | None = None,
    lazy: Literal[True] = True,
) -> _pl.LazyFrame: ...
@overload
def penguins(
    engine: Literal["dask"],
    *,
    engine_kwargs: dict[str, Any] | None = None,
    lazy: Literal[True] = True,
) -> _dd.DataFrame: ...
@overload
def earthquakes(
    engine: Literal["pandas"],
    *,
    engine_kwargs: dict[str, Any] | None = None,
    lazy: Literal[False] = False,
) -> _pd.DataFrame: ...
@overload
def earthquakes(
    engine: Literal["polars"],
    *,
    engine_kwargs: dict[str, Any] | None = None,
    lazy: Literal[False] = False,
) -> _pl.DataFrame: ...
@overload
def earthquakes(
    engine: Literal["polars"],
    *,
    engine_kwargs: dict[str, Any] | None = None,
    lazy: Literal[True] = True,
) -> _pl.LazyFrame: ...
@overload
def earthquakes(
    engine: Literal["dask"],
    *,
    engine_kwargs: dict[str, Any] | None = None,
    lazy: Literal[True] = True,
) -> _dd.DataFrame: ...
@overload
def apple_stocks(
    engine: Literal["pandas"],
    *,
    engine_kwargs: dict[str, Any] | None = None,
    lazy: Literal[False] = False,
) -> _pd.DataFrame: ...
@overload
def apple_stocks(
    engine: Literal["polars"],
    *,
    engine_kwargs: dict[str, Any] | None = None,
    lazy: Literal[False] = False,
) -> _pl.DataFrame: ...
@overload
def apple_stocks(
    engine: Literal["polars"],
    *,
    engine_kwargs: dict[str, Any] | None = None,
    lazy: Literal[True] = True,
) -> _pl.LazyFrame: ...
@overload
def apple_stocks(
    engine: Literal["dask"],
    *,
    engine_kwargs: dict[str, Any] | None = None,
    lazy: Literal[True] = True,
) -> _dd.DataFrame: ...
@overload
def stocks(
    engine: Literal["pandas"],
    *,
    engine_kwargs: dict[str, Any] | None = None,
    lazy: Literal[False] = False,
) -> _pd.DataFrame: ...
@overload
def stocks(
    engine: Literal["polars"],
    *,
    engine_kwargs: dict[str, Any] | None = None,
    lazy: Literal[False] = False,
) -> _pl.DataFrame: ...
@overload
def stocks(
    engine: Literal["polars"],
    *,
    engine_kwargs: dict[str, Any] | None = None,
    lazy: Literal[True] = True,
) -> _pl.LazyFrame: ...
@overload
def stocks(
    engine: Literal["dask"],
    *,
    engine_kwargs: dict[str, Any] | None = None,
    lazy: Literal[True] = True,
) -> _dd.DataFrame: ...
@overload
def nyc_taxi_remote(
    engine: Literal["pandas"],
    *,
    engine_kwargs: dict[str, Any] | None = None,
    lazy: Literal[False] = False,
) -> _pd.DataFrame: ...
@overload
def nyc_taxi_remote(
    engine: Literal["polars"],
    *,
    engine_kwargs: dict[str, Any] | None = None,
    lazy: Literal[False] = False,
) -> _pl.DataFrame: ...
@overload
def nyc_taxi_remote(
    engine: Literal["polars"],
    *,
    engine_kwargs: dict[str, Any] | None = None,
    lazy: Literal[True] = True,
) -> _pl.LazyFrame: ...
@overload
def nyc_taxi_remote(
    engine: Literal["dask"],
    *,
    engine_kwargs: dict[str, Any] | None = None,
    lazy: Literal[True] = True,
) -> _dd.DataFrame: ...
@overload
def air_temperature(
    engine: Literal["xarray"],
    *,
    engine_kwargs: dict[str, Any] | None = None,
    # lazy: Literal[False] = False,
) -> _xr.Dataset: ...
@overload
def synthetic_clusters(
    engine: Literal["pandas"],
    *,
    lazy: Literal[False] = False,
    total_points: int = 1_000_000,
) -> _pd.DataFrame: ...
@overload
def synthetic_clusters(
    engine: Literal["polars"],
    *,
    lazy: Literal[False] = False,
    total_points: int = 1_000_000,
) -> _pl.DataFrame: ...
@overload
def synthetic_clusters(
    engine: Literal["polars"],
    *,
    lazy: Literal[True] = True,
    total_points: int = 1_000_000,
) -> _pl.LazyFrame: ...
@overload
def synthetic_clusters(
    engine: Literal["dask"],
    *,
    lazy: bool = False,
    total_points: int = 1_000_000,
) -> _dd.DataFrame: ...
@overload
def landsat_rgb(
    engine: Literal["rioxarray"],
    *,
    engine_kwargs: dict[str, Any] | None = None,
) -> _xr.Dataset: ...
@overload
def us_states(
    engine: Literal["geopandas"],
    *,
    engine_kwargs: dict[str, Any] | None = None,
) -> _gpd.GeoDataFrame: ...



from __future__ import annotations

import importlib
import os
from functools import reduce
from pathlib import Path
from typing import Any
from urllib.parse import urlsplit

from platformdirs import user_cache_path

_DATAPATH = Path(__file__).parent / "_data"
_CACHEPATH = user_cache_path() / "hvsampledata"

_EAGER_TABULAR_LOOKUP = {
    "pandas": {"csv": "read_csv", "parquet": "read_parquet"},
    "polars": {"csv": "read_csv", "parquet": "read_parquet"},
}
_LAZY_TABULAR_LOOKUP = {
    "polars": {"csv": "scan_csv", "parquet": "scan_parquet"},
    "dask": {"csv": "dataframe.read_csv", "parquet": "dataframe.read_parquet"},
}
_EAGER_GRIDDED_LOOKUP = {
    "xarray": {"dataset": "open_dataset", "dataarray": "open_dataarray"},
}


def _get_path(dataset: str) -> Path:
    if dataset.startswith("http"):
        dataset_name = urlsplit(dataset).path.lstrip("/")
        path = _CACHEPATH / dataset_name
        if not path.exists():
            path.parent.mkdir(parents=True, exist_ok=True)
            _download_data(url=dataset, path=path)
        return path
    else:
        path = _DATAPATH / dataset
        return path


def _download_data(*, url, path):
    from urllib3 import PoolManager

    from hvsampledata import __version__

    headers = {"User-Agent": f"hvsampledata {__version__}"}

    http = PoolManager()
    response = None

    try:
        response = http.request("GET", url, preload_content=False, headers=headers)

        if response.status == 200:
            with open(path, "wb") as f:
                for chunk in response.stream(1024):
                    f.write(chunk)
            print(f"File saved to {path}")
        else:
            print(f"Failed to download file. HTTP Status: {response.status}")
    except Exception:
        print("Failed to download file")
        if path.exists():
            os.remove(path)
    finally:
        if response is not None:
            response.release_conn()
        http.clear()


def _get_method(*, engine: str | None, format: str, engine_lookups: dict[str, dict[str, str]]):
    # TODO: Should also work with .tar.gz like files
    if isinstance(engine, str):
        mod = importlib.import_module(engine)
        attr = engine_lookups[engine][format]
        if attr.count("."):
            importlib.import_module(".".join([engine, *attr.split(".")[:-1]]))
        return reduce(getattr, attr.split("."), mod)
    else:
        from importlib.util import find_spec

        for tab_engine in engine_lookups:
            if find_spec(tab_engine):
                return _get_method(engine=tab_engine, format=format, engine_lookups=engine_lookups)
        print("No available engines can be imported")


def _load_tabular(
    dataset: str,
    *,
    format: str | None = None,
    engine: str | None = None,
    engine_kwargs: dict[str, Any] | None = None,
    lazy: bool = False,
):
    path = _get_path(dataset)
    format = format or os.fspath(dataset).split(".")[-1]
    engine_lookup = _LAZY_TABULAR_LOOKUP if lazy else _EAGER_TABULAR_LOOKUP
    engine_function = _get_method(engine=engine, format=format, engine_lookups=engine_lookup)
    data = engine_function(path, **(engine_kwargs or {}))
    return data


def _load_gridded(
    dataset: str,
    *,
    format: str | None = None,
    engine: str | None = None,
    engine_kwargs: dict[str, Any] | None = None,
    # lazy=False,  # TODO: Add support for lazy
):
    path = _get_path(dataset)
    format = format or os.fspath(dataset).split(".")[-1]
    engine_lookup = _EAGER_GRIDDED_LOOKUP
    engine_function = _get_method(engine=engine, format=format, engine_lookups=engine_lookup)
    data = engine_function(path, **(engine_kwargs or {}))
    return data



from __future__ import annotations

import pytest

import hvsampledata as hvs
from hvsampledata._util import _EAGER_GRIDDED_LOOKUP

datasets = [hvs.air_temperature]


@pytest.mark.parametrize("dataset", datasets)
@pytest.mark.parametrize("engine", list(_EAGER_GRIDDED_LOOKUP))
def test_eager_load(dataset, engine):
    pytest.importorskip("xarray")
    pytest.importorskip("netCDF4")
    df = dataset(engine=engine)
    if engine == "xarray":
        import xarray as xr

        assert isinstance(df, xr.Dataset)
    else:
        msg = f"Not valid engine {engine}"
        raise ValueError(msg)


def test_air_temperature():
    pytest.importorskip("xarray")
    pytest.importorskip("netCDF4")
    import numpy as np

    ds = hvs.air_temperature("xarray")
    assert ds.air.shape == (20, 25, 53)
    assert ds.attrs == {
        "Conventions": "COARDS",
        "title": "4x daily NMC reanalysis (1948)",
        "description": "Data is from NMC initialized reanalysis\n(4x/day).  These are the 0.9950 sigma level values.",
        "platform": "Model",
        "references": "http://www.esrl.noaa.gov/psd/data/gridded/data.ncep.reanalysis.html",
    }
    assert ds.coords.dtypes == {
        "lat": np.dtype("float32"),
        "lon": np.dtype("float32"),
        "time": np.dtype("datetime64[ns]"),
    }
    assert ds.dtypes == {"air": np.dtype("float64")}
    assert not ds.air.isnull().any().item()


def test_landsat_rgb():
    pytest.importorskip("rioxarray")
    import numpy as np

    ds = hvs.landsat_rgb("rioxarray")

    assert ds.rgb.shape == (3, 215, 237)
    assert str(ds.rio.crs) == "EPSG:32618"
    assert ds.coords.dtypes == {
        "band": np.dtype("int64"),
        "x": np.dtype("float64"),
        "y": np.dtype("float64"),
        "spatial_ref": np.dtype("int64"),
    }
    assert ds.rgb.dtype == np.dtype("uint8")
    assert not ds.rgb.isnull().any().item()


def test_penguins_rgba():
    pytest.importorskip("xarray")
    import numpy as np

    ds = hvs.penguins_rgba("xarray")

    assert ds.rgba.shape == (100, 100, 4)
    assert ds.coords.dtypes == {
        "y": np.dtype("int64"),
        "x": np.dtype("int64"),
        "channel": np.dtype("<U1"),
    }
    assert ds.rgba.dtype == np.dtype("uint8")
    assert not ds.rgba.isnull().any().item()
